import pandas as pd

from snp_flip.reference_genome import get_reference_genome_data

from snp_flip.bim_file import read_bim_file

def create_snp_table(bim_file, fa_file):

    bim_table = read_bim_file(bim_file)

    reference_genome_data = get_reference_genome_data(bim_table, fa_file)

    snp_table = pd.concat([bim_table, reference_genome_data], axis=1)

    return snp_table


def find_strand(bim_table):

    forward_mask = (bim_table["reference"] == bim_table["allele_1"]) | \
                   (bim_table["reference"] == bim_table["allele_2"])

    reverse_mask = (bim_table["reference_rev"] == bim_table["allele_1"]) | \
                   (bim_table["reference_rev"] == bim_table["allele_2"])

    ambiguous_mask = (forward_mask & reverse_mask) | \
                     (~(forward_mask | reverse_mask))

    strand_series = bim_table["reference"].copy()
    strand_series[forward_mask] = "forward"
    strand_series[reverse_mask] = "reverse"
    strand_series[ambiguous_mask] = "ambiguous"
    strand_series.name = "strand"

    return pd.concat([bim_table, strand_series], axis=1)



from __future__ import print_function

import re
from sys import stderr

from operator import itemgetter

import pandas as pd

from pyfaidx import Fasta



def get_reference_genome_data(bim_table, genome_fasta):

    pyfaidx_genome = Fasta(genome_fasta, as_raw=True, read_ahead=100000)
    genome_dict = convert_fa_chromosome_names(pyfaidx_genome)

    _warn_about_bim_chromos_not_in_fa(bim_table, genome_dict)

    overlapping_chromosomes = set(genome_dict.keys()).intersection(set(bim_table.chromosome.drop_duplicates()))
    genome_dict = {c: d for c, d in genome_dict.items() if c in overlapping_chromosomes}

    chromosome_data = []
    for chromosome, nucleotides in genome_dict.items():

        positions = bim_table[bim_table["chromosome"] == chromosome]["position"] - 1
        _get_snps = itemgetter(*list(positions))

        snp_nucleotides = [snp.upper() for snp in _get_snps(str(nucleotides))]
        chromosome_data.append(pd.DataFrame(snp_nucleotides))

        _check_for_N(snp_nucleotides, chromosome)

    genome_data = pd.concat(chromosome_data)
    genome_data.columns = ["reference"]
    genome_data["reference_rev"] = genome_data["reference"].apply(
        lambda n: {"A": "T",
                   "T": "A",
                   "C": "G",
                   "G": "C"}[n])


    return genome_data.reset_index(drop=True)


def _warn_about_bim_chromos_not_in_fa(bim_table, genome_dict):

    for chromosome in bim_table["chromosome"].drop_duplicates():
        if chromosome not in genome_dict:
            print("Chromosome {} in .bim, but not in fasta file.".format(
                chromosome), file=stderr)


def _check_for_N(snps, chromosome):

    nb_ns = sum([1 for snp in snps if snp == "N"])
    if nb_ns != 0:
        print("There were {nb_ns} 'N' nucleotides in chromosome"
              "{chromosome}.".format(**vars()), file=stderr)


def convert_fa_chromosome_names(fasta_dict):

    return {_convert_fa_name(k): fasta_dict[k] for k in fasta_dict.keys()
            if _is_valid_chromosome(k)}


def _is_valid_chromosome(name):

    for c in "_.":
        if c in name.split(None, 1)[0]:
            return False

    return True


def _convert_fa_name(name):

    bim_name = re.compile(r"(chr)?(?P<chr_symbol>\d+|X|Y|M)(?:T)?")
    result = bim_name.match(name).group("chr_symbol")

    return result



from itertools import groupby

import pandas as pd


def read_bim_file(bim_file):
    bim_table = pd.read_table(bim_file, sep=r"\s+", header=None,
                              names=["chromosome", "snp_name",
                                     "genetic_distance", "position",
                                     "allele_1", "allele_2"],
                              dtype={"chromosome": str})

    _check_if_rows_grouped_by_chromosomes(bim_table["chromosome"])

    return bim_table


def _check_if_rows_grouped_by_chromosomes(chromosomes):


    nb_chromosomes = len(chromosomes.drop_duplicates())
    uniq_chromosomes = len(list(groupby(chromosomes, lambda x: x)))

    if nb_chromosomes != uniq_chromosomes:
        raise ValueError("The rows in the bim file are not grouped by " \
                         "chromosomes. Exiting.")
